import fs from 'fs'
import { applyOverridesInPlace } from './overrides'

export interface HanziWriterData {
  strokes: string[]
  medians: number[][][]
  radStrokes?: number[]
}

const overrides: Record<string, Partial<HanziWriterData>> = {
  个: {
    // this was incorrectly set to "radStrokes":[2]
    radStrokes: [0, 1]
  },
  顛: animDataConcat(
    animDataSlice('颠', 0, 10),
    animDataSlice('順', 3, 12)
  )
}

export async function queryHanziWriterData (
  char: string
): Promise<HanziWriterData | undefined> {
  const hanziWriterJsonPath = getHanziWriterJsonPath(char)
  if (!fs.existsSync(hanziWriterJsonPath)) {
    return undefined
  }
  const data = JSON.parse(
    await fs.promises.readFile(hanziWriterJsonPath, 'utf-8')
  ) as HanziWriterData
  return applyOverridesInPlace(data, overrides[char])
}

function loadHanziWriterDataWithoutOverridesSync (
  char: string
): HanziWriterData {
  const hanziWriterJsonPath = getHanziWriterJsonPath(char)
  if (!fs.existsSync(hanziWriterJsonPath)) {
    throw new Error(`${char} not found`)
  }
  const data = JSON.parse(
    fs.readFileSync(hanziWriterJsonPath, 'utf-8')
  ) as HanziWriterData
  return data
}

function getHanziWriterJsonPath (char: string): string {
  return `node_modules/hanzi-writer-data/${char}.json`
}

function animDataSlice (
  char: string,
  fromIncl: number,
  toExcl: number
): Partial<HanziWriterData> {
  const { strokes, radStrokes, medians } =
    loadHanziWriterDataWithoutOverridesSync(char)
  const animData: Partial<HanziWriterData> = {
    strokes: strokes.slice(fromIncl, toExcl),
    medians: medians.slice(fromIncl, toExcl)
  }
  if (radStrokes?.every(i => i >= fromIncl && i < toExcl) === true) {
    animData.radStrokes = radStrokes.map(i => i - fromIncl)
  }
  return animData
}

function animDataConcat (
  left: Partial<HanziWriterData>,
  right: Partial<HanziWriterData>
): Partial<HanziWriterData> {
  const {
    strokes: leftStrokes,
    radStrokes: leftRadStrokes,
    medians: leftMedians
  } = left
  const {
    strokes: rightStrokes,
    radStrokes: rightRadStrokes,
    medians: rightMedians
  } = right
  const together: Partial<HanziWriterData> = {}
  if (leftStrokes !== undefined && rightStrokes !== undefined) {
    together.strokes = leftStrokes.concat(rightStrokes)
  }
  if (leftRadStrokes !== undefined && rightRadStrokes !== undefined) {
    throw new Error('How can both sides have a radical?')
  } else if (leftRadStrokes !== undefined) {
    together.radStrokes = [...leftRadStrokes]
  } else if (rightRadStrokes !== undefined) {
    if (leftStrokes === undefined) {
      throw new Error('no strokes on the left')
    }
    together.radStrokes = rightRadStrokes.map(i => i + leftStrokes.length)
  }
  if (leftMedians !== undefined && rightMedians !== undefined) {
    together.medians = leftMedians.concat(rightMedians)
  }
  return together
}
