import {
  type StrokeType,
  type HanziData,
  type TradOverrides
} from './hanzi-data'
import cnchar from 'cnchar'
import cncharOrder from 'cnchar-order'
import cncharRadical from 'cnchar-radical'
import cncharTrad from 'cnchar-trad'
import { getTraditionalInfo } from './traditional'
import {
  type HanziWriterData,
  queryHanziWriterData
} from './query-hanzi-writer'
import { applyOverridesInPlace } from './overrides'

cnchar.use(cncharOrder, cncharRadical, cncharTrad)

export interface QueryOpts {
  readonly char: string
}

const noHanziWriterData: HanziWriterData = { strokes: [], medians: [] }
const overrides: Record<string, Partial<HanziData>> = {
  艸: {
    strokeTypes: [
      '𠃊',
      '㇑',
      '㇓',
      '𠃊',
      '㇑',
      '㇑'
    ]
  }
}

export async function query ({ char }: QueryOpts): Promise<HanziData> {
  if (char.length !== 1) {
    throw new Error(`Can only query single chars, got: ${char}`)
  }
  const data = await queryDefault(char)
  data.trad = await queryTraditional(char, data)
  return data
}

async function queryDefault (char: string): Promise<HanziData> {
  const hanziWriterData = (await queryHanziWriterData(char)) ??
    noHanziWriterData
  const totalStrokeCount = hanziWriterData.strokes.length
  const simplified = cnchar.radical(char)[0]
  const radical = simplified.radical
  const extra = Math.max(0, totalStrokeCount - simplified.radicalCount)
  const count = `${simplified.radicalCount}+${extra}`
  const strokeTypes = queryShortenedCncharStrokeTypes(char, totalStrokeCount)
  return applyOverridesInPlace(
    {
      radical,
      count,
      strokeTypes,
      ...hanziWriterData,
      trad: {}
    },
    overrides[char]
  )
}

/**
 * Query stroke shapes for the characters and return them in a shortened form.
 *
 * If the stroke count does not match the given expected count, return an empty
 * array instead.
 */
function queryShortenedCncharStrokeTypes (
  char: string,
  expectedLen: number
): StrokeType[] {
  const strokeShapes = cnchar.stroke(char, 'order', 'shape')[0]
  if (!Array.isArray(strokeShapes) || strokeShapes.length === 0) {
    console.error(`stroke table data unavailable for ${char}`)
    return []
  }
  if (strokeShapes.length !== expectedLen) {
    console.error(`${char}: ${
      strokeShapes.length
    } strokes in cnchar, but ${
      expectedLen
    } in hanzi writer, omitting cnchar stroke shapes`)
    return []
  }
  // if cnchar has multiple stroke types, only keep first
  return strokeShapes.map(t => t.split('|')[0])
}

async function queryTraditional (
  char: string,
  defaultData: HanziData
): Promise<TradOverrides> {
  const overrides: TradOverrides = {}
  const traditional = await getTraditionalInfo(char)
  for (const prop of ['radical', 'count']) {
    if (traditional[prop] !== defaultData[prop]) {
      overrides[prop] = traditional[prop]
    }
  }
  return overrides
}
